use std::{
	sync::mpsc::{self, Sender},
	thread,
};

use fltk::{
	app,
	enums::{Color, FrameType},
	misc::Chart,
	prelude::{GroupExt, WidgetExt, WindowExt},
	window::Window,
};
use fltk_theme::{color_themes, ColorTheme};
use rand::Rng;

const SIZE: u32 = 50;
const MAX_VALUE: f64 = 100.0;

type Sorter = fn(SortVisualizer);

#[derive(Debug, Clone)]
struct SortVisualizer {
	chart: Chart,
	arr: Vec<f64>,
	sorter: Sorter,
	tx: Sender<SortVisualizer>,
}

impl SortVisualizer {
	fn new(tx: Sender<SortVisualizer>, arr: Vec<f64>, sorter: Sorter) -> Self {
		let mut chart = Chart::default();
		chart.set_size(500, 250);
		chart.set_frame(FrameType::FlatBox);
		chart.set_bounds(0.0, MAX_VALUE);
		Self {
			tx,
			chart,
			arr,
			sorter,
		}
	}

	fn sort(self) {
		(self.sorter)(self);
	}

	fn redraw(&mut self) {
		self.chart.clear();
		for i in &self.arr {
			self.chart.add(*i, "", Color::Cyan);
		}
	}
}

fn quick_sort(arr: &mut Vec<f64>, vis: &mut SortVisualizer) {}

fn bubble_sort(mut vis: SortVisualizer) {
	let n = vis.arr.len();

	for i in 0..n {
		for j in 0..n - i - 1 {
			if vis.arr[j] > vis.arr[j + 1] {
				vis.arr.swap(j, j + 1);

				vis.tx.send(vis.clone()).unwrap();
			}
		}
	}
}

fn main() {
	let mut arr: Vec<f64> = (1..SIZE)
		.map(|_| rand::rng().random_range(1.0..MAX_VALUE))
		.collect();

	let (tx, rx) = mpsc::channel();

	let sorts = [bubble_sort];
	// let mut visualizers = vec![];

	let app = app::App::default().with_scheme(app::Scheme::Gtk);
	let theme = ColorTheme::new(&color_themes::fleet::DARK1);
	theme.apply();

	let mut window = Window::default()
		.with_size(640, 480)
		.with_label("FLTK Visualizer");
	window.size_range(640, 480, 0, 0);

	let mut vis = SortVisualizer::new(tx, arr, bubble_sort);
	// visualizers.push(vis);

	window.end();
	window.show();

	let handle = thread::spawn(move || {
		vis.sort();
	});

	app::add_idle3(move |_| {
		let vis = rx.recv();
		if let Ok(mut vis) = vis {
			vis.redraw();
		}

		window.redraw();
		app::sleep(0.016);
	});

	app.run().unwrap();

	handle.join().unwrap();
}
